from .arcade import ArcadeProvider
from .base import IAliasProvider, IAreaProvider, ICurveProvider, IItemListProvider, IPlayerProvider, IRegionProvider, IScoreProvider, ISongProvider
from .divingfish import DivingFishProvider
from .hybrid import HybridProvider
from .local import LocalProvider
from .lxns import LXNSProvider
from .wechat import WechatProvider
from .yuzu import YuzuProvider

__all__ = [
    "IAliasProvider",
    "IAreaProvider",
    "ICurveProvider",
    "IItemListProvider",
    "IPlayerProvider",
    "IRegionProvider",
    "IScoreProvider",
    "ISongProvider",
    "ArcadeProvider",
    "DivingFishProvider",
    "HybridProvider",
    "LocalProvider",
    "LXNSProvider",
    "WechatProvider",
    "YuzuProvider",
]



import asyncio
import hashlib
from typing import TYPE_CHECKING

from maimai_py.models import *

from .base import ISongProvider
from .divingfish import DivingFishProvider
from .lxns import LXNSProvider

if TYPE_CHECKING:
    from maimai_py.maimai import MaimaiClient


class HybridProvider(ISongProvider):
    """The provider that fetches data from the LXNS and DivingFish, and hybrids them together.

    This provider is used to provide full notes and utage songs data, taking advantage of the LXNS and DivingFish APIs.

    LXNS: https://maimai.lxns.net/

    DivingFish: https://www.diving-fish.com/maimaidx/prober/
    """

    base_url_lxns = LXNSProvider.base_url
    base_url_divingfish = DivingFishProvider.base_url

    def _hash(self) -> str:
        return hashlib.md5(b"hybrid").hexdigest()

    @staticmethod
    def _deser_note(diff: dict, key: str) -> int:
        if "notes" in diff:
            if "is_buddy" in diff and diff["is_buddy"]:
                return diff["notes"]["left"][key] + diff["notes"]["right"][key]
            return diff["notes"][key]
        return 0

    @staticmethod
    def _deser_diff(difficulty: dict, id: int, dv_dict: dict) -> SongDifficulty:
        diff = SongDifficulty(
            type=SongType[difficulty["type"].upper()],
            level=difficulty["level"],
            level_value=difficulty["level_value"],
            level_index=LevelIndex(difficulty["difficulty"]),
            note_designer=difficulty["note_designer"],
            version=difficulty["version"],
            tap_num=HybridProvider._deser_note(difficulty, "tap"),
            hold_num=HybridProvider._deser_note(difficulty, "hold"),
            slide_num=HybridProvider._deser_note(difficulty, "slide"),
            touch_num=HybridProvider._deser_note(difficulty, "touch"),
            break_num=HybridProvider._deser_note(difficulty, "break"),
            curve=None,
        )
        dv_id = diff._get_divingfish_id(id)
        if dv_difficulty := dv_dict.get(dv_id, None):
            if diff.level_index.value < len(dv_difficulty["charts"]):
                notes: list = dv_difficulty["charts"][diff.level_index.value]["notes"]
                diff.tap_num = notes[0] if len(notes) > 0 else 0
                diff.hold_num = notes[1] if len(notes) > 1 else 0
                diff.slide_num = notes[2] if len(notes) > 2 else 0
                diff.touch_num = notes[3] if len(notes) > 3 else 0
                diff.break_num = notes[4] if len(notes) > 4 else 0
        return diff

    async def get_songs(self, client: "MaimaiClient") -> list[Song]:
        lxns, divingfish = await asyncio.gather(
            asyncio.create_task(client._client.get(self.base_url_lxns + "api/v0/maimai/song/list")),
            asyncio.create_task(client._client.get(self.base_url_divingfish + "music_data")),
        )
        lxns.raise_for_status()
        divingfish.raise_for_status()
        resp_lxns, resp_divingfish = lxns.json(), divingfish.json()

        dv_songs = {int(song["id"]): song for song in resp_divingfish}
        unique_songs: dict[int, Song] = {}

        for song in resp_lxns["songs"]:
            id = int(song["id"]) % 10000
            if id not in unique_songs:
                unique_songs[id] = LXNSProvider._deser_song(song)
            difficulties = unique_songs[id].difficulties
            difficulties.standard.extend(HybridProvider._deser_diff(diff, id, dv_songs) for diff in song["difficulties"].get("standard", []))
            difficulties.dx.extend(HybridProvider._deser_diff(diff, id, dv_songs) for diff in song["difficulties"].get("dx", []))
            difficulties.utage.extend(LXNSProvider._deser_diff_utage(diff) for diff in song["difficulties"].get("utage", []))
        return list(unique_songs.values())



from abc import abstractmethod
from typing import TYPE_CHECKING

from maimai_py.models import *

if TYPE_CHECKING:
    from maimai_py.maimai import MaimaiClient


class ISongProvider:
    """The provider that fetches songs from a specific source.

    Available providers: `DivingFishProvider`, `LXNSProvider`
    """

    @abstractmethod
    def _hash(self) -> str: ...

    @abstractmethod
    async def get_songs(self, client: "MaimaiClient") -> list[Song]:
        """@private"""
        raise NotImplementedError()


class IAliasProvider:
    """The provider that fetches song aliases from a specific source.

    Available providers: `YuzuProvider`, `LXNSProvider`
    """

    @abstractmethod
    def _hash(self) -> str: ...

    @abstractmethod
    async def get_aliases(self, client: "MaimaiClient") -> list[SongAlias]:
        """@private"""
        raise NotImplementedError()


class IPlayerProvider:
    """The provider that fetches players from a specific source.

    Available providers: `DivingFishProvider`, `LXNSProvider`
    """

    @abstractmethod
    def _hash(self) -> str: ...

    @abstractmethod
    async def get_player(self, identifier: PlayerIdentifier, client: "MaimaiClient") -> Player:
        """@private"""
        raise NotImplementedError()


class IScoreProvider:
    """The provider that fetches scores from a specific source.

    Available providers: `DivingFishProvider`, `LXNSProvider`, `WechatProvider`
    """

    @abstractmethod
    def _hash(self) -> str: ...

    @abstractmethod
    async def get_scores(self, identifier: PlayerIdentifier, client: "MaimaiClient") -> list[Score]:
        """@private"""
        raise NotImplementedError()

    @abstractmethod
    async def update_scores(self, identifier: PlayerIdentifier, scores: list[Score], client: "MaimaiClient") -> None:
        """@private"""
        raise NotImplementedError()


class ICurveProvider:
    """The provider that fetches statistics curves from a specific source.

    Available providers: `DivingFishProvider`
    """

    @abstractmethod
    def _hash(self) -> str: ...

    @abstractmethod
    async def get_curves(self, client: "MaimaiClient") -> dict[tuple[int, SongType], list[CurveObject | None]]:
        """@private"""
        raise NotImplementedError()


class IRegionProvider:
    """The provider that fetches player regions from a specific source.

    Available providers: `ArcadeProvider`
    """

    @abstractmethod
    def _hash(self) -> str: ...

    @abstractmethod
    async def get_regions(self, identifier: PlayerIdentifier, client: "MaimaiClient") -> list[PlayerRegion]:
        """@private"""
        raise NotImplementedError()


class IItemListProvider:
    """The provider that fetches player item list data from a specific source.

    Available providers: `LXNSProvider`, `LocalProvider`
    """

    @abstractmethod
    def _hash(self) -> str: ...

    @abstractmethod
    async def get_icons(self, client: "MaimaiClient") -> dict[int, PlayerIcon]:
        """@private"""
        raise NotImplementedError()

    @abstractmethod
    async def get_nameplates(self, client: "MaimaiClient") -> dict[int, PlayerNamePlate]:
        """@private"""
        raise NotImplementedError()

    @abstractmethod
    async def get_frames(self, client: "MaimaiClient") -> dict[int, PlayerFrame]:
        """@private"""
        raise NotImplementedError()

    @abstractmethod
    async def get_partners(self, client: "MaimaiClient") -> dict[int, PlayerPartner]:
        """@private"""
        raise NotImplementedError()

    @abstractmethod
    async def get_charas(self, client: "MaimaiClient") -> dict[int, PlayerChara]:
        """@private"""
        raise NotImplementedError()

    @abstractmethod
    async def get_trophies(self, client: "MaimaiClient") -> dict[int, PlayerTrophy]:
        """@private"""
        raise NotImplementedError()


class IAreaProvider:
    """The provider that fetches area data from a specific source.

    Available providers: `LocalProvider`
    """

    @abstractmethod
    def _hash(self) -> str: ...

    @abstractmethod
    async def get_areas(self, lang: str, client: "MaimaiClient") -> dict[str, Area]:
        """@private"""
        raise NotImplementedError()



from dataclasses import dataclass
from datetime import datetime
from typing import Any, MutableMapping, Sequence

from maimai_py.enums import *
from maimai_py.exceptions import *
from maimai_py.utils import UNSET, _UnsetSentinel


@dataclass(slots=True)
class Song:
    id: int
    title: str
    artist: str
    genre: Genre
    bpm: int
    map: str | None
    version: int
    rights: str | None
    aliases: list[str] | None
    disabled: bool
    difficulties: "SongDifficulties"

    def get_difficulty(self, type: SongType, level_index: LevelIndex | None) -> "SongDifficulty | None":
        if type == SongType.DX:
            return next((diff for diff in self.difficulties.dx if diff.level_index == level_index), None)
        if type == SongType.STANDARD:
            return next((diff for diff in self.difficulties.standard if diff.level_index == level_index), None)
        if type == SongType.UTAGE:
            return next(iter(self.difficulties.utage), None)


@dataclass(slots=True)
class SongDifficulties:
    standard: list["SongDifficulty"]
    dx: list["SongDifficulty"]
    utage: list["SongDifficultyUtage"]

    def _get_children(self, song_type: SongType | _UnsetSentinel = UNSET) -> Sequence["SongDifficulty"]:
        if song_type == UNSET:
            return self.standard + self.dx + self.utage
        return self.dx if song_type == SongType.DX else self.standard if song_type == SongType.STANDARD else self.utage


@dataclass(slots=True)
class CurveObject:
    sample_size: int
    fit_level_value: float
    avg_achievements: float
    stdev_achievements: float
    avg_dx_score: float
    rate_sample_size: dict[RateType, int]
    fc_sample_size: dict[FCType, int]


@dataclass(slots=True)
class SongDifficulty:
    type: SongType
    level: str
    level_value: float
    level_index: LevelIndex
    note_designer: str
    version: int
    tap_num: int
    hold_num: int
    slide_num: int
    touch_num: int
    break_num: int
    curve: CurveObject | None

    def _get_divingfish_id(self, id: int) -> int:
        if id < 0 or id > 9999:
            raise ValueError("Invalid song ID")
        if self.type == SongType.DX:
            return id + 10000
        elif self.type == SongType.UTAGE:
            return id + 100000
        return id


@dataclass(slots=True)
class SongDifficultyUtage(SongDifficulty):
    kanji: str
    description: str
    is_buddy: bool


@dataclass(slots=True)
class SongAlias:
    """@private"""

    song_id: int
    aliases: list[str]


@dataclass(slots=True)
class PlayerIdentifier:
    qq: int | None = None
    username: str | None = None
    friend_code: int | None = None
    credentials: str | MutableMapping[str, str] | None = None

    def __post_init__(self):
        if self.qq is None and self.username is None and self.friend_code is None and self.credentials is None:
            raise InvalidPlayerIdentifierError("At least one of the following must be provided: qq, username, friend_code, credentials")

    def _as_diving_fish(self) -> dict[str, Any]:
        if self.qq:
            return {"qq": str(self.qq)}
        elif self.username:
            return {"username": self.username}
        elif self.friend_code:
            raise InvalidPlayerIdentifierError("Friend code is not applicable for Diving Fish")
        else:
            raise InvalidPlayerIdentifierError("No valid identifier provided")

    def _as_lxns(self) -> str:
        if self.friend_code:
            return str(self.friend_code)
        elif self.qq:
            return f"qq/{str(self.qq)}"
        elif self.username:
            raise InvalidPlayerIdentifierError("Username is not applicable for LXNS")
        else:
            raise InvalidPlayerIdentifierError("No valid identifier provided")


@dataclass(slots=True)
class ArcadeResponse:
    """@private"""

    errno: int | None = None
    errmsg: str | None = None
    data: dict[str, Any] | bytes | list[Any] | None = None

    @staticmethod
    def _raise_for_error(resp: "ArcadeResponse") -> None:
        if resp.errno and resp.errno != 0:
            if resp.errno > 1000:
                raise ArcadeError(resp.errmsg)
            elif resp.errno > 100:
                raise TitleServerError(resp.errmsg)
            elif resp.errno > 0:
                raise AimeServerError(resp.errmsg)


@dataclass(slots=True)
class PlayerItem:
    @staticmethod
    def _namespace() -> str:
        raise NotImplementedError


@dataclass(slots=True)
class PlayerTrophy(PlayerItem):
    id: int
    name: str
    color: str

    @staticmethod
    def _namespace():
        return "trophies"


@dataclass(slots=True)
class PlayerIcon(PlayerItem):
    id: int
    name: str
    description: str | None = None
    genre: str | None = None

    @staticmethod
    def _namespace():
        return "icons"


@dataclass(slots=True)
class PlayerNamePlate(PlayerItem):
    id: int
    name: str
    description: str | None = None
    genre: str | None = None

    @staticmethod
    def _namespace():
        return "nameplates"


@dataclass(slots=True)
class PlayerFrame(PlayerItem):
    id: int
    name: str
    description: str | None = None
    genre: str | None = None

    @staticmethod
    def _namespace():
        return "frames"


@dataclass(slots=True)
class PlayerPartner(PlayerItem):
    id: int
    name: str

    @staticmethod
    def _namespace():
        return "partners"


@dataclass(slots=True)
class PlayerChara(PlayerItem):
    id: int
    name: str

    @staticmethod
    def _namespace():
        return "charas"


@dataclass(slots=True)
class PlayerRegion:
    region_id: int
    region_name: str
    play_count: int
    created_at: datetime


@dataclass(slots=True)
class Player:
    name: str
    rating: int


@dataclass(slots=True)
class DivingFishPlayer(Player):
    nickname: str
    plate: str
    additional_rating: int


@dataclass(slots=True)
class LXNSPlayer(Player):
    friend_code: int
    course_rank: int
    class_rank: int
    star: int
    frame: PlayerFrame | None
    icon: PlayerIcon | None
    trophy: PlayerTrophy | None
    name_plate: PlayerNamePlate | None
    upload_time: str


@dataclass(slots=True)
class ArcadePlayer(Player):
    is_login: bool
    icon: PlayerIcon | None
    trophy: PlayerTrophy | None
    name_plate: PlayerNamePlate | None


@dataclass(slots=True)
class AreaCharacter:
    name: str
    illustrator: str
    description1: str
    description2: str
    team: str
    props: dict[str, str]


@dataclass(slots=True)
class AreaSong:
    id: int | None
    title: str
    artist: str
    description: str
    illustrator: str | None
    movie: str | None


@dataclass(slots=True)
class Area:
    id: str
    name: str
    comment: str
    description: str
    video_id: str
    characters: list[AreaCharacter]
    songs: list[AreaSong]


@dataclass(slots=True)
class Score:
    id: int
    level: str
    level_index: LevelIndex
    achievements: float | None
    fc: FCType | None
    fs: FSType | None
    dx_score: int | None
    dx_rating: float | None
    rate: RateType
    type: SongType

    def _compare(self, other: "Score | None") -> "Score":
        if other is None:
            return self
        if self.dx_score != other.dx_score:  # larger value is better
            return self if (self.dx_score or 0) > (other.dx_score or 0) else other
        if self.achievements != other.achievements:  # larger value is better
            return self if (self.achievements or 0) > (other.achievements or 0) else other
        if self.rate != other.rate:  # smaller value is better
            self_rate = self.rate.value if self.rate is not None else 100
            other_rate = other.rate.value if other.rate is not None else 100
            return self if self_rate < other_rate else other
        if self.fc != other.fc:  # smaller value is better
            self_fc = self.fc.value if self.fc is not None else 100
            other_fc = other.fc.value if other.fc is not None else 100
            return self if self_fc < other_fc else other
        if self.fs != other.fs:  # bigger value is better
            self_fs = self.fs.value if self.fs is not None else -1
            other_fs = other.fs.value if other.fs is not None else -1
            return self if self_fs > other_fs else other
        return self  # we consider they are equal


@dataclass(slots=True)
class PlateScore:
    id: int
    level_index: LevelIndex
    achievements: float | None
    fc: FCType | None
    fs: FSType | None
    rate: RateType
    type: SongType

    @staticmethod
    def _from_score(score: Score) -> "PlateScore":
        return PlateScore(
            id=score.id,
            level_index=score.level_index,
            achievements=score.achievements,
            fc=score.fc,
            fs=score.fs,
            rate=score.rate,
            type=score.type,
        )

    def _join(self, other: "PlateScore | None") -> "PlateScore":
        if other is not None:
            if self.level_index != other.level_index or self.type != other.type:
                raise ValueError("Cannot join scores with different level indexes or types")
            self.achievements = max(self.achievements or 0, other.achievements or 0)
            if self.fc != other.fc:
                self_fc = self.fc.value if self.fc is not None else 100
                other_fc = other.fc.value if other.fc is not None else 100
                selected_value = min(self_fc, other_fc)
                self.fc = FCType(selected_value) if selected_value != 100 else None
            if self.fs != other.fs:
                self_fs = self.fs.value if self.fs is not None else -1
                other_fs = other.fs.value if other.fs is not None else -1
                selected_value = max(self_fs, other_fs)
                self.fs = FSType(selected_value) if selected_value != -1 else None
            if self.rate != other.rate:
                selected_value = min(self.rate.value, other.rate.value)
                self.rate = RateType(selected_value)
        return self


@dataclass(slots=True)
class PlateSong:
    id: int
    title: str
    artist: str
    levels: list[LevelIndex]

    @staticmethod
    def _from_song(song: Song | None, song_type: SongType, exclude_remaster: bool = False) -> "PlateSong | None":
        if song:
            levels = [diff.level_index for diff in song.difficulties._get_children(song_type)]
            if exclude_remaster and LevelIndex.ReMASTER in levels:
                levels.remove(LevelIndex.ReMASTER)
            return PlateSong(id=song.id, title=song.title, artist=song.artist, levels=levels)

    def _as_empty(self) -> "PlateSong":
        return PlateSong(id=self.id, title=self.title, artist=self.artist, levels=[])

    def _as_full(self) -> "PlateSong":
        return PlateSong(id=self.id, title=self.title, artist=self.artist, levels=self.levels.copy())


@dataclass(slots=True)
class PlateObject:
    song: PlateSong
    scores: list[PlateScore]



import asyncio
from dataclasses import dataclass
from importlib.util import find_spec
from typing import Annotated, Callable, Literal
from urllib.parse import unquote, urlparse

from maimai_py import ArcadeProvider, DivingFishProvider, LXNSProvider, MaimaiClient, MaimaiPlates, MaimaiScores, MaimaiSongs
from maimai_py.models import *

PlateAttrs = Literal["remained", "cleared", "played", "all"]


@dataclass(slots=True)
class ScorePublic(Score):
    song_name: str
    level_value: float


@dataclass(slots=True)
class PlayerBests:
    rating: int
    rating_b35: int
    rating_b15: int
    scores_b35: list[ScorePublic]
    scores_b15: list[ScorePublic]


@dataclass(slots=True)
class ParsedQRCode:
    credentials: str


def pagination(page_size, page, data):
    total_pages = (len(data) + page_size - 1) // page_size
    if page < 1 or page > total_pages:
        return []

    start = (page - 1) * page_size
    end = page * page_size
    return data[start:end]


def xstr(s: str | None) -> str:
    return "" if s is None else str(s).lower()


def istr(i: list | None) -> str:
    return "" if i is None else "".join(i).lower()


def get_filters(functions: dict[Any, Callable[..., bool]]):
    union = [flag for cond, flag in functions.items() if cond is not None]
    filter = lambda obj: all([flag(obj) for flag in union])
    return filter


async def ser_score(score: Score, songs: dict[int, Song]) -> ScorePublic | None:
    if (song := songs.get(score.id)) and (diff := song.get_difficulty(score.type, score.level_index)):
        return ScorePublic(
            id=score.id,
            song_name=song.title,
            level=score.level,
            level_index=score.level_index,
            level_value=diff.level_value,
            achievements=score.achievements,
            fc=score.fc,
            fs=score.fs,
            dx_score=score.dx_score,
            dx_rating=score.dx_rating,
            rate=score.rate,
            type=score.type,
        )


async def ser_bests(maimai_scores: MaimaiScores, maimai_songs: MaimaiSongs) -> PlayerBests:
    song_ids = [score.id for score in maimai_scores.scores_b35 + maimai_scores.scores_b15]
    songs: list[Song] = await maimai_songs.get_batch(song_ids) if len(song_ids) > 0 else []
    required_songs: dict[int, Song] = {song.id: song for song in songs}
    async with asyncio.TaskGroup() as tg:
        b35_tasks = [tg.create_task(ser_score(score, required_songs)) for score in maimai_scores.scores_b35]
        b15_tasks = [tg.create_task(ser_score(score, required_songs)) for score in maimai_scores.scores_b15]
    scores_b35, scores_b15 = [v for task in b35_tasks if (v := task.result())], [v for task in b15_tasks if (v := task.result())]
    return PlayerBests(
        rating=maimai_scores.rating,
        rating_b35=maimai_scores.rating_b35,
        rating_b15=maimai_scores.rating_b15,
        scores_b35=scores_b35,
        scores_b15=scores_b15,
    )


if find_spec("fastapi"):
    from fastapi import APIRouter, Depends, FastAPI, Header, Query, Request
    from fastapi.openapi.utils import get_openapi
    from fastapi.responses import JSONResponse

    def dep_lxns_player(friend_code: int | None = None, qq: int | None = None):
        return PlayerIdentifier(qq=qq, friend_code=friend_code)

    def dep_diving_player_auth(username: str | None = None, credentials: str | None = None):
        return PlayerIdentifier(username=username, credentials=credentials)

    def dep_diving_player(username: str | None = None, qq: int | None = None):
        return PlayerIdentifier(qq=qq, username=username)

    def dep_arcade_player(credentials: str):
        return PlayerIdentifier(credentials=credentials)

    class MaimaiRoutes:
        _client: MaimaiClient

        _lxns_token: str | None = None
        _divingfish_token: str | None = None
        _arcade_proxy: str | None = None

        def __init__(
            self,
            client: MaimaiClient,
            lxns_token: str | None = None,
            divingfish_token: str | None = None,
            arcade_proxy: str | None = None,
        ):
            self._client = client
            self._lxns_token = lxns_token
            self._divingfish_token = divingfish_token
            self._arcade_proxy = arcade_proxy

        async def _get_songs(
            self,
            id: int | None = None,
            title: str | None = None,
            artist: str | None = None,
            genre: Genre | None = None,
            bpm: int | None = None,
            map: str | None = None,
            version: int | None = None,
            type: SongType | None = None,
            level: str | None = None,
            versions: Version | None = None,
            keywords: str | None = None,
            page: int = Query(1, ge=1),
            page_size: int = Query(100, ge=1, le=1000),
        ):
            songs: MaimaiSongs = await self._client.songs()
            type_func: Callable[[Song], bool] = lambda song: song.difficulties._get_children(type) != []  # type: ignore
            level_func: Callable[[Song], bool] = lambda song: any([diff.level == level for diff in song.difficulties._get_children()])
            versions_func: Callable[[Song], bool] = lambda song: versions.value <= song.version < all_versions[all_versions.index(versions) + 1].value  # type: ignore
            keywords_func: Callable[[Song], bool] = lambda song: xstr(keywords) in xstr(song.title) + xstr(song.artist) + istr(song.aliases)
            filters = get_filters({type: type_func, level: level_func, versions: versions_func, keywords: keywords_func})
            results = [x async for x in songs.filter(id=id, title=title, artist=artist, genre=genre, bpm=bpm, map=map, version=version) if filters(x)]
            return pagination(page_size, page, results)

        async def _get_icons(
            self,
            id: int | None = None,
            name: str | None = None,
            description: str | None = None,
            genre: str | None = None,
            keywords: str | None = None,
            page: int = Query(1, ge=1),
            page_size: int = Query(100, ge=1, le=1000),
        ):
            items = await self._client.items(PlayerIcon)
            if id is not None:
                return [item] if (item := items.by_id(id)) else []
            keyword_func: Callable[[PlayerIcon], bool] = lambda icon: xstr(keywords) in (xstr(icon.name) + xstr(icon.description) + xstr(icon.genre))
            filters = get_filters({keywords: keyword_func})
            results = [x async for x in items.filter(name=name, description=description, genre=genre) if filters(x)]
            return pagination(page_size, page, results)

        async def _get_nameplates(
            self,
            id: int | None = None,
            name: str | None = None,
            description: str | None = None,
            genre: str | None = None,
            keywords: str | None = None,
            page: int = Query(1, ge=1),
            page_size: int = Query(100, ge=1, le=1000),
        ):
            items = await self._client.items(PlayerNamePlate)
            if id is not None:
                return [item] if (item := items.by_id(id)) else []
            keyword_func: Callable[[PlayerNamePlate], bool] = lambda icon: xstr(keywords) in (
                xstr(icon.name) + xstr(icon.description) + xstr(icon.genre)
            )
            filters = get_filters({keywords: keyword_func})
            results = [x async for x in items.filter(name=name, description=description, genre=genre) if filters(x)]
            return pagination(page_size, page, results)

        async def _get_frames(
            self,
            id: int | None = None,
            name: str | None = None,
            description: str | None = None,
            genre: str | None = None,
            keywords: str | None = None,
            page: int = Query(1, ge=1),
            page_size: int = Query(100, ge=1, le=1000),
        ):
            items = await self._client.items(PlayerFrame)
            if id is not None:
                return [item] if (item := items.by_id(id)) else []
            keyword_func: Callable[[PlayerFrame], bool] = lambda icon: xstr(keywords) in (xstr(icon.name) + xstr(icon.description) + xstr(icon.genre))
            filters = get_filters({keywords: keyword_func})
            results = [x async for x in items.filter(name=name, description=description, genre=genre) if filters(x)]
            return pagination(page_size, page, results)

        async def _get_trophies(
            self,
            id: int | None = None,
            name: str | None = None,
            color: str | None = None,
            keywords: str | None = None,
            page: int = Query(1, ge=1),
            page_size: int = Query(100, ge=1, le=1000),
        ):
            items = await self._client.items(PlayerTrophy)
            if id is not None:
                return [item] if (item := items.by_id(id)) else []
            keyword_func: Callable[[PlayerTrophy], bool] = lambda icon: xstr(keywords) in (xstr(icon.name) + xstr(icon.color))
            filters = get_filters({keywords: keyword_func})
            results = [x async for x in items.filter(name=name, color=color) if filters(x)]
            return pagination(page_size, page, results)

        async def _get_charas(
            self,
            id: int | None = None,
            name: str | None = None,
            keywords: str | None = None,
            page: int = Query(1, ge=1),
            page_size: int = Query(100, ge=1, le=1000),
        ):
            items = await self._client.items(PlayerChara)
            if id is not None:
                return [item] if (item := items.by_id(id)) else []
            results = items.filter(name=name or keywords)
            return pagination(page_size, page, results)

        async def _get_partners(
            self,
            id: int | None = None,
            name: str | None = None,
            keywords: str | None = None,
            page: int = Query(1, ge=1),
            page_size: int = Query(100, ge=1, le=1000),
        ):
            items = await self._client.items(PlayerPartner)
            if id is not None:
                return [item] if (item := items.by_id(id)) else []
            results = items.filter(name=name or keywords)
            return pagination(page_size, page, results)

        async def _get_areas(
            self,
            lang: Literal["ja", "zh"] = "ja",
            id: str | None = None,
            name: str | None = None,
            page: int = Query(1, ge=1),
            page_size: int = Query(100, ge=1, le=1000),
        ):
            areas = await self._client.areas(lang)
            if id is not None:
                return [area] if (area := await areas.by_id(id)) else []
            if name is not None:
                return [area] if (area := await areas.by_name(name)) else []
            return pagination(page_size, page, await areas.get_all())

        async def _get_player_lxns(self, player: PlayerIdentifier = Depends(dep_lxns_player)):
            provider = LXNSProvider(self._lxns_token)
            return await self._client.players(player, provider)

        async def _get_player_diving(self, player: PlayerIdentifier = Depends(dep_diving_player)):
            provider = DivingFishProvider(self._divingfish_token)
            return await self._client.players(player, provider)

        async def _get_player_arcade(self, player: PlayerIdentifier = Depends(dep_arcade_player)):
            provider = ArcadeProvider(self._arcade_proxy)
            return await self._client.players(player, provider)

        async def _get_scores_lxns(self, player: PlayerIdentifier = Depends(dep_lxns_player)):
            provider = LXNSProvider(self._lxns_token)
            scores = await self._client.scores(player, provider=provider)
            return scores.scores  # no pagination because it costs more

        async def _get_scores_diving(self, player: PlayerIdentifier = Depends(dep_diving_player)):
            provider = DivingFishProvider(self._divingfish_token)
            scores = await self._client.scores(player, provider=provider)
            return scores.scores  # no pagination because it costs more

        async def _get_scores_arcade(self, player: PlayerIdentifier = Depends(dep_arcade_player)):
            provider = ArcadeProvider(self._arcade_proxy)
            scores = await self._client.scores(player, provider=provider)
            return scores.scores  # no pagination because it costs more

        async def _update_scores_lxns(self, scores: list[Score], player: PlayerIdentifier = Depends(dep_lxns_player)):
            provider = LXNSProvider(self._lxns_token)
            await self._client.updates(player, scores, provider=provider)

        async def _update_scores_diving(self, scores: list[Score], player: PlayerIdentifier = Depends(dep_diving_player_auth)):
            provider = DivingFishProvider(self._divingfish_token)
            await self._client.updates(player, scores, provider=provider)

        async def _get_bests_lxns(self, player: PlayerIdentifier = Depends(dep_lxns_player)):
            provider = LXNSProvider(self._lxns_token)
            songs, scores = await asyncio.gather(
                asyncio.create_task(self._client.songs()),
                asyncio.create_task(self._client.scores(player, provider=provider)),
            )
            return await ser_bests(scores, songs)

        async def _get_bests_diving(self, player: PlayerIdentifier = Depends(dep_diving_player)):
            provider = DivingFishProvider(self._divingfish_token)
            songs, scores = await asyncio.gather(
                asyncio.create_task(self._client.songs()),
                asyncio.create_task(self._client.scores(player, provider=provider)),
            )
            return await ser_bests(scores, songs)

        async def _get_bests_arcade(self, player: PlayerIdentifier = Depends(dep_arcade_player)):
            provider = ArcadeProvider(self._arcade_proxy)
            songs, scores = await asyncio.gather(
                asyncio.create_task(self._client.songs()),
                asyncio.create_task(self._client.scores(player, provider=provider)),
            )
            return await ser_bests(scores, songs)

        async def _get_plate_lxns(
            self, plate: str, attr: Literal["remained", "cleared", "played", "all"] = "remained", player: PlayerIdentifier = Depends(dep_lxns_player)
        ):
            provider = LXNSProvider(self._lxns_token)
            plates: MaimaiPlates = await self._client.plates(player, plate, provider=provider)
            return await getattr(plates, f"get_{attr}")()

        async def _get_plate_diving(self, plate: str, attr: PlateAttrs = "remained", player: PlayerIdentifier = Depends(dep_diving_player)):
            provider = DivingFishProvider(self._divingfish_token)
            plates: MaimaiPlates = await self._client.plates(player, plate, provider=provider)
            return await getattr(plates, f"get_{attr}")()

        async def _get_plate_arcade(self, plate: str, attr: PlateAttrs = "remained", player: PlayerIdentifier = Depends(dep_arcade_player)):
            provider = ArcadeProvider(self._arcade_proxy)
            plates: MaimaiPlates = await self._client.plates(player, plate, provider=provider)
            return await getattr(plates, f"get_{attr}")()

        async def _get_region_arcade(self, player: PlayerIdentifier = Depends(dep_arcade_player)):
            provider = ArcadeProvider(self._arcade_proxy)
            return await self._client.regions(player, provider=provider)

        async def _get_qrcode_arcade(self, qrcode: str):
            identifier = await self._client.qrcode(qrcode, http_proxy=self._arcade_proxy)
            return ParsedQRCode(credentials=str(identifier.credentials))

        def get_base(self) -> APIRouter:
            router = APIRouter()
            router.add_api_route(
                "/songs",
                self._get_songs,
                name="get_songs",
                methods=["GET"],
                response_model=list[Song],
                description="Get songs by various filters, filters are combined by AND",
            )
            router.add_api_route(
                "/icons",
                self._get_icons,
                name="get_icons",
                methods=["GET"],
                response_model=list[PlayerIcon],
                description="Get player icons by various filters, filters are combined by AND",
            )
            router.add_api_route(
                "/nameplates",
                self._get_nameplates,
                name="get_nameplates",
                methods=["GET"],
                response_model=list[PlayerNamePlate],
                description="Get player nameplates by various filters, filters are combined by AND",
            )
            router.add_api_route(
                "/frames",
                self._get_frames,
                name="get_frames",
                methods=["GET"],
                response_model=list[PlayerFrame],
                description="Get player frames by various filters, filters are combined by AND",
            )
            router.add_api_route(
                "/trophies",
                self._get_trophies,
                name="get_trophies",
                methods=["GET"],
                response_model=list[PlayerTrophy],
                description="Get player trophies by various filters, filters are combined by AND",
            )
            router.add_api_route(
                "/charas",
                self._get_charas,
                name="get_charas",
                methods=["GET"],
                response_model=list[PlayerChara],
                description="Get player charas by various filters, filters are combined by AND",
            )
            router.add_api_route(
                "/partners",
                self._get_partners,
                name="get_partners",
                methods=["GET"],
                response_model=list[PlayerPartner],
                description="Get player partners by various filters, filters are combined by AND",
            )
            router.add_api_route(
                "/areas",
                self._get_areas,
                name="get_areas",
                methods=["GET"],
                response_model=list[Area],
                description="Get areas",
            )
            return router

        def get_divingfish(self) -> APIRouter:
            router = APIRouter()
            router.add_api_route(
                "/players",
                self._get_player_diving,
                name="get_player_diving",
                methods=["GET"],
                response_model=DivingFishPlayer,
                description="Get player info from Diving Fish",
            )
            router.add_api_route(
                "/scores",
                self._get_scores_diving,
                name="get_scores_diving",
                methods=["GET"],
                response_model=list[Score],
                description="Get player ALL scores from Diving Fish",
            )
            router.add_api_route(
                "/scores",
                self._update_scores_diving,
                name="update_scores_diving",
                methods=["POST"],
                description="Update player scores to Diving Fish, should provide the user's username and password, or import token as credentials.",
            )
            router.add_api_route(
                "/bests",
                self._get_bests_diving,
                name="get_bests_diving",
                methods=["GET"],
                response_model=PlayerBests,
                description="Get player b50 scores from Diving Fish",
            )
            router.add_api_route(
                "/plates",
                self._get_plate_diving,
                name="get_plate_diving",
                methods=["GET"],
                response_model=list[PlateObject],
                description="Get player plates from Diving Fish",
            )
            return router

        def get_lxns(self) -> APIRouter:
            router = APIRouter()
            router.add_api_route(
                "/players",
                self._get_player_lxns,
                name="get_player_lxns",
                methods=["GET"],
                response_model=LXNSPlayer,
                description="Get player info from LXNS",
            )
            router.add_api_route(
                "/scores",
                self._get_scores_lxns,
                name="get_scores_lxns",
                methods=["GET"],
                response_model=list[Score],
                description="Get player ALL scores from LXNS",
            )
            router.add_api_route(
                "/scores",
                self._update_scores_lxns,
                name="update_scores_lxns",
                methods=["POST"],
                description="Update player scores to LXNS",
            )
            router.add_api_route(
                "/bests",
                self._get_bests_lxns,
                name="get_bests_lxns",
                methods=["GET"],
                response_model=PlayerBests,
                description="Get player b50 scores from LXNS",
            )
            router.add_api_route(
                "/plates",
                self._get_plate_lxns,
                name="get_plate_lxns",
                methods=["GET"],
                response_model=list[PlateObject],
                description="Get player plates from LXNS",
            )
            return router

        def get_arcade(self) -> APIRouter:
            router = APIRouter()
            router.add_api_route(
                "/players",
                self._get_player_arcade,
                name="get_player_arcade",
                methods=["GET"],
                response_model=ArcadePlayer,
                description="Get player info from Arcade",
            )
            router.add_api_route(
                "/scores",
                self._get_scores_arcade,
                name="get_scores_arcade",
                methods=["GET"],
                response_model=list[Score],
                description="Get player ALL scores from Arcade",
            )
            router.add_api_route(
                "/bests",
                self._get_bests_arcade,
                name="get_bests_arcade",
                methods=["GET"],
                response_model=PlayerBests,
                description="Get player b50 scores from Arcade",
            )
            router.add_api_route(
                "/plates",
                self._get_plate_arcade,
                name="get_plate_arcade",
                methods=["GET"],
                response_model=list[PlateObject],
                description="Get player plates from Arcade",
            )
            router.add_api_route(
                "/regions",
                self._get_region_arcade,
                name="get_region_arcade",
                methods=["GET"],
                response_model=list[PlayerRegion],
                description="Get player regions from Arcade",
            )
            router.add_api_route(
                "/qrcode",
                self._get_qrcode_arcade,
                name="get_qrcode_arcade",
                methods=["GET"],
                response_model=ParsedQRCode,
                description="Get encrypted player credentials from QR code",
            )
            return router


if all([find_spec(p) for p in ["fastapi", "uvicorn", "typer"]]):
    import typer
    import uvicorn
    from fastapi import APIRouter, Depends, FastAPI, Query, Request
    from fastapi.openapi.utils import get_openapi
    from fastapi.responses import JSONResponse

    # prepare for ASGI app
    asgi_app = FastAPI(title="maimai.py API", description="The definitive python wrapper for MaimaiCN related development.")
    maimai_routes = MaimaiRoutes(MaimaiClient())  # type: ignore

    # register routes and middlewares
    asgi_app.include_router(maimai_routes.get_base(), tags=["base"])
    asgi_app.include_router(maimai_routes.get_divingfish(), prefix="/divingfish", tags=["divingfish"])
    asgi_app.include_router(maimai_routes.get_lxns(), prefix="/lxns", tags=["lxns"])
    asgi_app.include_router(maimai_routes.get_arcade(), prefix="/arcade", tags=["arcade"])

    def main(
        host: Annotated[str, typer.Option(help="The host address to bind to.")] = "127.0.0.1",
        port: Annotated[int, typer.Option(help="The port number to bind to.")] = 8000,
        redis: Annotated[str | None, typer.Option(help="Redis server address, for example: redis://localhost:6379/0.")] = None,
        lxns_token: Annotated[str | None, typer.Option(help="LXNS token for LXNS API.")] = None,
        divingfish_token: Annotated[str | None, typer.Option(help="Diving Fish token for Diving Fish API.")] = None,
        arcade_proxy: Annotated[str | None, typer.Option(help="HTTP proxy for Arcade API.")] = None,
    ):
        # prepare for redis cache backend
        redis_backend = UNSET
        if redis and find_spec("redis"):
            from aiocache import RedisCache
            from aiocache.serializers import PickleSerializer

            redis_url = urlparse(redis)
            redis_backend = RedisCache(
                serializer=PickleSerializer(),
                endpoint=unquote(redis_url.hostname or "localhost"),
                port=redis_url.port or 6379,
                password=redis_url.password,
                db=int(unquote(redis_url.path).replace("/", "")),
            )

        # override the default maimai.py client
        maimai_client = MaimaiClient(cache=redis_backend)
        maimai_routes._client = maimai_client
        maimai_routes._lxns_token = lxns_token
        maimai_routes._divingfish_token = divingfish_token
        maimai_routes._arcade_proxy = arcade_proxy

        @asgi_app.exception_handler(MaimaiPyError)
        async def exception_handler(request: Request, exc: MaimaiPyError):
            return JSONResponse(
                status_code=400,
                content={"message": f"Oops! There goes a maimai.py error {exc}.", "details": repr(exc)},
            )

        @asgi_app.get("/", include_in_schema=False)
        async def root():
            return {"message": "Hello, maimai.py! Check /docs for more information."}

        # run the ASGI app with uvicorn
        uvicorn.run(asgi_app, host=host, port=port)

    def openapi():
        specs = get_openapi(
            title=asgi_app.title,
            version=asgi_app.version,
            openapi_version=asgi_app.openapi_version,
            description=asgi_app.description,
            routes=asgi_app.routes,
        )
        with open(f"openapi.json", "w") as f:
            json.dump(specs, f)

    if __name__ == "__main__":
        typer.run(main)


if find_spec("maimai_ffi") and find_spec("nuitka"):
    import json

    import cryptography
    import cryptography.fernet
    import cryptography.hazmat.backends
    import cryptography.hazmat.primitives.ciphers
    import maimai_ffi
    import maimai_ffi.model
    import maimai_ffi.request
    import redis
